#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2022 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################


import wx
import wx.grid as grd
import wx.propgrid as pg
from pathlib import Path
from dataclasses import dataclass
from typing import Optional, List, Callable
import datetime as dt
import threading
import numpy as np
import h5py
import csv
from PIL import ImageGrab, Image, ImageFile
import matplotlib as mpl
mpl.use('WXAgg')  # switch backend (case insensitive)
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as NavigationToolbar
from nardascripting.base.signalsharkdev import *
from nardascripting.base.usrscriptbase import *
from nardascripting.base.measthread import *
from nardascripting.base.scpi.enums import *
from .lldialogsbase import *
from ..lldata import *
from .. import filehandling as fh
from .images import scriptimages as ico


class RefMeasImportDlg(RefMeasSettingsBaseDlg):
    """Settings dialog"""

    def __init__(self, parent, filename: str, spectrum_settings: SpectrumSettings, ll_settings: LLSettingsData):
        """

        :param parent: wx.Window parent
        :param filename: The name of the HDF5 file.
        :param spectrum_settings: The current spectrum settings
        :param ll_settings: The limit line settings
        """
        super().__init__(parent)
        self.filename = filename
        self.spectrum_settings = spectrum_settings
        self.ll_settings = ll_settings
        self.tasks = {}

        self.cbx_ll_type_selection.Clear()
        ll_types = []
        for item in LimitTypes:
            name = str(item)
            if name == "BOTH":
                continue
            ll_types.append(name)
        self.cbx_ll_type_selection.AppendItems(ll_types)
        self.cbx_ll_type_selection.SetSelection(0)
        self.my_ok_btnOK.Enable(False)

    def on_dlg_show(self, event):
        """On form is shown"""
        msg = self.get_tasks_and_traces()
        if msg:
            with wx.MessageDialog(self, msg, 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
            self.Close()
        event.Skip()

    def get_tasks_and_traces(self) -> str:
        """Get all available task and trace names

        :return: Returns error message or empty string.
        """
        try:
            self.cbx_task_selection.Clear()
            task_items = ["Please select..."]
            self.cbx_trace_selection.Clear()
            self.tasks = {}
            with h5py.File(str(self.filename), 'r') as f:
                for task_id in f:
                    trace_list = []
                    task = f.get(task_id)
                    task_type = str(task.attrs["Task type"][0])[2:-1]
                    task_name = str(task).split('"')[1][1:]
                    if task_type == "Spectrum" or task_type == "RT Spectrum":
                        task_items.append(task_name)
                        spectrum_meas = task.get("Spectrum")
                        if spectrum_meas:
                            h5_trace_list = list(spectrum_meas.dtype.names)
                            for trace_type in h5_trace_list:
                                trace_list.append(TRACE_NAMES_H5.get(trace_type, "???"))
                        self.tasks[task_name] = trace_list
            self.cbx_task_selection.AppendItems(task_items)
            self.cbx_task_selection.SetSelection(0)
            return ""
        except Exception as e:
            return str(e)

    def on_task_selection(self, event):
        """Updates the trace list.

        :param event: Task selection from combo box in the dialog
        """
        try:
            selected_task = self.cbx_task_selection.GetStringSelection()
            if "Please select" in selected_task:
                return
            self.cbx_trace_selection.Clear()
            self.cbx_trace_selection.Enable(False)
            trace_list = self.tasks.get(selected_task)
            if trace_list is not None:
                self.cbx_trace_selection.AppendItems(trace_list)
                self.cbx_trace_selection.Enable(True)
            self.cbx_trace_selection.SetFocus()
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_trace_selection(self, event):
        """Updates the OK button.

        :param event: Trace selection from combo box in the dialog
        """
        try:
            self.my_ok_btnOK.Enable(True)
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_ok(self, event):
        """On OK button pressed.

        :param event: The OK button event.
        """
        try:
            selected_ll_type = self.cbx_ll_type_selection.GetStringSelection()
            if not selected_ll_type:
                return

            selected_task = self.cbx_task_selection.GetStringSelection()
            if not selected_task:
                return

            selected_trace = self.cbx_trace_selection.GetSelection()
            if selected_trace < 0:
                return

            self.ll_settings.ll_items.clear()
            ll_items: List[LLItem] = []

            with h5py.File(str(self.filename), 'r') as f:
                for task_id in f:
                    task = f.get(task_id)
                    task_type = str(task.attrs["Task type"][0]).replace("\'", "")
                    task_name = str(task).split('"')[1].replace('/', '')
                    if task_name != selected_task:
                        continue

                    spectrum_meas = task.get("Spectrum")
                    if spectrum_meas:
                        # Get unit and spectrum frequency values:
                        unit = str(spectrum_meas.attrs["Level unit"])[3:-2]
                        spec_unit = str(self.spectrum_settings.unit)
                        if spec_unit != unit:
                            msg = "Unit does not match HDF5: \'{}\', Device: \'{}\'!".format(unit, spec_unit)
                            with wx.MessageDialog(self, msg, 'Error',
                                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                                dlg.ShowModal()
                                return
                        bin_count = spectrum_meas.attrs["Number of bins"][0]
                        binwidth = spectrum_meas.attrs["Binwidth / Hz"][0]
                        fstart = spectrum_meas.attrs["Frequency of first bin / Hz"][0]

                        index = 0
                        values = spectrum_meas[()]
                        for item in values:
                            freq = float(fstart + (index * binwidth))
                            lev = float(item[selected_trace])
                            if selected_ll_type == "UPPER":
                                ll_items.append(LLItem(freq, lev, -999.0))
                            else:
                                ll_items.append(LLItem(freq, 999.0, lev))
                            index += 1

                        if len(ll_items) == bin_count:
                            self.ll_settings.ll_items = ll_items
                    break
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()


class AddLevOffset(AddLevOffsetBase):
    """AddLevOffset dialog"""

    def __init__(self, parent, limit_type: LimitTypes):
        """Initialization of dialog.

        :param parent: wx.Window parent
        :param limit_type: Type of limit to add offset.
        """
        super().__init__(parent)
        self.cbx_limit_type_selection.Clear()
        limit_type = str(limit_type)
        limit_types = []
        for item in LimitTypes:
            name = str(item)
            if name == "BOTH":
                continue
            limit_types.append(name)
        self.cbx_limit_type_selection.AppendItems(limit_types)
        if limit_type in  limit_types:
            self.cbx_limit_type_selection.SetStringSelection(limit_type)
        else:
            self.cbx_limit_type_selection.SetSelection(0)
        self.limit_type: LimitTypes = LimitTypes.from_string(limit_type)
        self.offset: float = 0.0

    def on_sctr_offset_mouse_wheel(self, event):
        """Upper limit value mouse wheel"""
        try:
            value = self.sctr_offset.GetValue()
            inc = self.sctr_offset.GetIncrement()
            wheel_rotation = event.GetWheelRotation()
            if wheel_rotation > 0:
                value += inc
            else:
                value -= inc
            self.sctr_offset.SetValue(value)
        except:
            pass
        finally:
            event.Skip()

    def __apply_value(self):
        """Method that applies user input to property"""
        try:
            selected_limit_type = self.cbx_limit_type_selection.GetStringSelection()
            if not selected_limit_type:
                self.SetReturnCode(wx.CANCEL)
                return

            self.limit_type = LimitTypes.from_string(selected_limit_type)
            self.offset = self.sctr_offset.GetValue()
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()

    def on_sctr_offset_enter_key(self, event):
        """Key down event"""
        try:
            self.__apply_value()
            self.EndModal(wx.ID_OK)
        except:
            pass
        finally:
            event.Skip()

    def on_ok(self, event):
        """OK button event"""
        try:
            self.__apply_value()
        except:
            pass
        finally:
            event.Skip()


class NumValueDlg(NumValueDlgBase):
    """NumValue dialog"""

    def __init__(self, parent, value: float = 0.0, label: str = "Value", unit: str = ""):
        """Initialization of dialog.

        :param parent: wx.Window parent
        """
        super().__init__(parent)
        self.lbl_designator.SetLabelText(label)
        self.sctr_value.SetValue(value)
        self.lbl_unit.SetLabelText(unit)

    def GetValue(self):
        """Gets the current value"""
        return self.sctr_value.GetValue()

    def SetValue(self, value: float):
        """Sets the current value"""
        self.sctr_value.SetValue(value)

    def on_sctr_value_mouse_wheel(self, event):
        """Upper limit value mouse wheel"""
        try:
            value = self.sctr_value.GetValue()
            inc = self.sctr_value.GetIncrement()
            wheel_rotation = event.GetWheelRotation()
            if wheel_rotation > 0:
                value += inc
            else:
                value -= inc
            self.sctr_value.SetValue(value)
        except:
            pass
        finally:
            event.Skip()

    def on_sctr_value_enter_key(self, event):
        """Key down event"""
        try:
            self.EndModal(wx.ID_OK)
        except:
            pass
        finally:
            event.Skip()


class TracePlotCanvasPnl(wx.Panel):
    """wxWidget plot panel based on matplotlib plots"""

    def __init__(self, parent):
        """ wxWidget plot panel initialization with matplotlib features

        :param parent: The parent widget.
        :type parent: wx.Widget
        """
        super(TracePlotCanvasPnl, self).__init__(parent, size=(450, 360))
        # self.tag = tag
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvas(self, -1, self.figure)

        # Workaround for size issue
        self.canvas.SetMinSize(wx.Size(10, 10))

        # self.toolbar = NavigationToolbar(self.canvas)
        # self.toolbar.Realize()

        self.llc_freq = []
        self.llc_lev_upper = []
        self.llc_lev_lower = []

        self.x_values = []
        self.lines = {}

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.canvas, proportion=1, flag=wx.ALL | wx.GROW)
        # self.sizer.Add(self.toolbar, 0, wx.LEFT | wx.EXPAND)
        self.SetSizer(self.sizer)
        self.Fit()
        self.trace_colors = {
            "": "#000000",
            "UNKNOWN": "#000000",
            "RMS": "#00C000",
            "PPk": "#FF0000",
            "MPk": "#0000FF",
            "Avg": "#0000FF",
            "Smp": "#0000FF",
            "MnR": "#003955",
            "AvR": "#FF00FF",
            "MxR": "#553900",
            "MxP": "#AA0000",
            "MnP": "#0000AA",
            "MxA": "#0000AA",
            "MxS": "#0000AA"
        }

        self.yaxis_label = {
            LevelUnits.UNKNOWN: "Unknown",
            LevelUnits.dBm: "Level / dBm",
            LevelUnits.dBV: "Level / dBV",
            LevelUnits.dBmV: "Level / dBmV",
            LevelUnits.dBuV: "Level / dBµV",
            LevelUnits.W_m2: "Level / W/m²",
            LevelUnits.W_cm2: "Level / mW/cm²",
            LevelUnits.A_m: "Level / A/m",
            LevelUnits.V_m: "Level / V/m",
            LevelUnits.dBA_m: "Level / dBA/m",
            LevelUnits.dBV_m: "Level / dBV/m",
            LevelUnits.dBmV_m: "Level / dBmV/m",
            LevelUnits.dBuV_m: "Level / dBµV/m"
        }

    def clear_memory(self):
        """Clears the actual traces"""
        self.axes.cla()
        self.lines = {}
        self.x_values = []

    def update_settings(self, spectrum_settings: SpectrumSettings,
                        ll_settings: LLSettingsData, title: str = ""):
        """ Updates plot settings according to the selected limitline type

        :param spectrum_settings: Measurement settings of the Spectrum View
        :param ll_settings: Limit line settings
        :param title: Title of the plot panel
        :return: Returns None
        """
        self.clear_memory()

        unit = spectrum_settings.unit
        y_label_str = self.yaxis_label.get(unit, "Unknown")
        self.axes.set_ylabel(y_label_str)
        self.axes.set_xlabel("Frequency/MHz")
        self.axes.set_title(title)
        self.axes.set_autoscale_on(False)
        self.axes.set_autoscalex_on(False)
        self.axes.set_autoscaley_on(False)
        self.axes.grid(True)
        f1 = spectrum_settings.fstart / 1e6
        f2 = spectrum_settings.fstop / 1e6
        self.axes.set_xbound(f1, f2)
        min_y = spectrum_settings.lmin
        max_y = spectrum_settings.lmax
        plot_ll = False

        # Get limit line data:
        if ll_settings.ll_type == LimitLineTypes.HORIZONTAL:
            plot_ll = True
            if ll_settings.limit_type == LimitTypes.UPPER or ll_settings.limit_type == LimitTypes.BOTH:
                if max_y < ll_settings.upper_limit_value:
                    max_y = ll_settings.upper_limit_value + 1

            if ll_settings.limit_type == LimitTypes.LOWER or ll_settings.limit_type == LimitTypes.BOTH:
                if min_y > ll_settings.lower_limit_value:
                    min_y = ll_settings.lower_limit_value - 1

            self.llc_freq = [f1, f2]
            self.llc_lev_upper = [ll_settings.upper_limit_value, ll_settings.upper_limit_value]
            self.llc_lev_lower = [ll_settings.lower_limit_value, ll_settings.lower_limit_value]
        else:
            factor = 1/1e6
            self.llc_freq = []
            self.llc_lev_upper = []
            self.llc_lev_lower = []
            if ll_settings.ll_items is not None and len(ll_settings.ll_items) > 1:
                plot_ll = True
                for item in ll_settings.ll_items:
                    self.llc_freq.append(item.frequency * factor)
                    self.llc_lev_upper.append(item.upper_level)
                    self.llc_lev_lower.append(item.lower_level)

                if ll_settings.limit_type == LimitTypes.UPPER or ll_settings.limit_type == LimitTypes.BOTH:
                    max_upper_value = max(self.llc_lev_upper)
                    if max_y < max_upper_value:
                        max_y = max_upper_value + 1

                if ll_settings.limit_type == LimitTypes.LOWER or ll_settings.limit_type == LimitTypes.BOTH:
                    min_lower_value = min(self.llc_lev_lower)
                    if min_y > min_lower_value:
                        min_y = min_lower_value - 1

        unit_str = str(unit)
        if "dB" in unit_str:
            self.axes.set_yscale('linear')
        else:
            self.axes.set_yscale('log')
        self.axes.set_ybound(min_y, max_y)

        self.figure.subplots_adjust(left=0.2, right=0.9, bottom=0.2, top=0.9)
        self.x_values = spectrum_settings.get_flist(1e6)

        # Plotting upper limitline
        if plot_ll:
            if ll_settings.limit_type == LimitTypes.UPPER or ll_settings.limit_type == LimitTypes.BOTH:
                self.lines['UPPER'] = self.axes.plot(self.llc_freq, self.llc_lev_upper, label='UPPER', color='orange')

            # Plotting lower limitline
            if ll_settings.limit_type == LimitTypes.LOWER or ll_settings.limit_type == LimitTypes.BOTH:
                self.lines['LOWER'] = self.axes.plot(self.llc_freq, self.llc_lev_lower, label='LOWER', color='purple')

        # Plotting empty trace
        y_values = [0.0] * spectrum_settings.bin_count
        trace_name = str(ll_settings.trace)
        self.lines[trace_name] = self.axes.plot(self.x_values, y_values,
                                                     color=self.trace_colors.get(trace_name, "#000000"),
                                                     label=trace_name)

        # Legend Settings
        self.axes.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
                         ncol=4, mode="expand", borderaxespad=0.)

    def update_y_values(self, trace: str, y_values: list):
        """ Updates a trace with new y data.
        :param trace: Trace name
        :type trace: str
        :param y_values: List of level values.
        :type y_values: list of float
        """
        line = self.lines.get(trace)
        if not line:
            return
        line[0].set_ydata(y_values)

    def update_plot(self):
        self.canvas.draw()


class LLEdit(LLSettingsBaseDlg):
    """Limitline settings editor dialog"""

    def __init__(self, parent, spectrum_settings: SpectrumSettings, trace_data: dict, current_working_dir: str):
        """Limitline settings editor dialog

        :param parent: wx.Widgets parent.
        :param spectrum_settings: Spectrum settings for displaying Traces.
        :param trace_data: Current trace data readout from the device.
        :param current_working_dir: The current working directory readout from the device.
        """
        super().__init__(parent)

        self.spectrum_settings = spectrum_settings
        # load settings:
        self.ll_settings = fh.load_dclass_from_xml(LLSettingsData, fh.LL_LAST_SETTINGS_FILENAME)
        if self.ll_settings is None:
            self.ll_settings = LLSettingsData()
        self.trace_data = trace_data
        self.current_working_dir = current_working_dir

        self.custom_tbl_csv = self.ll_settings.custom_tbl_csv
        self.custom_tbl_h5 = self.ll_settings.custom_tbl_h5
        self.ignore_tbl_events = False

        # Trigger settings:
        # -----------------------------------------------------------
        self.pg_trigger_settings.SetCellBackgroundColour(wx.Colour( 51, 51, 51 ))
        self.pg_trigger_settings.SetCellBackgroundColour(wx.Colour(51, 51, 51))
        self.pg_trigger_settings.SetCellTextColour(wx.Colour( 255, 255, 255 ))
        self.pg_trigger_settings.SetCellDisabledTextColour(wx.Colour( 102, 102, 102 ))
        self.pg_trigger_settings.SetLineColour(wx.Colour( 102, 102, 102 ))
        self.pg_trigger_settings.SetCaptionBackgroundColour(wx.Colour(80, 80, 80))
        self.pg_trigger_settings.SetCaptionTextColour(wx.Colour( 255, 255, 255 ))
        self.pg_trigger_settings.SetEmptySpaceColour(wx.Colour(51, 51, 51))
        self.pg_trigger_settings.SetMarginColour(wx.Colour(51, 51, 51))

        self.pgi_cat_trigger_event = self.pg_trigger_settings.Append(
            pg.PropertyCategory(u"Trigger event", u"Trigger event"))
        delay_values = list(range(1, 60))
        delay_values.insert(0, 0.5)
        delay_values_str = [str(x) for x in delay_values]
        self.pgi_trigger_delay = pg.EnumProperty(label=u"Trigger delay (s)", name=u"Trigger delay",
                                                 labels=delay_values_str, values=delay_values, value=0)
        self.pgi_trigger_delay = self.pg_trigger_settings.Append(self.pgi_trigger_delay)
        self.pgi_trigger_on_ll = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Trigger on limit line event", u"Trigger on limit line event"))
        self.pgi_trigger_on_button = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Trigger on button press", u"Trigger on button press"))
        self.pgi_button_key_code = self.pg_trigger_settings.Append(
            pg.StringProperty(u"Button key code", u"Button key code"))

        self.pgi_cat_sound_display_settings = self.pg_trigger_settings.Append(
            pg.PropertyCategory(u"Sound and Display Settings", u"Sound and Display Settings"))
        self.pgi_sound_on_button_trigger = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Sound on button press", u"Sound on button press"))
        self.pgi_sound_on_ll_trigger = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Sound on limit line event", u"limit line event"))

        meas_screen_types_str = []
        meas_screen_types_val = []
        for item in MeasScreenTypes:
            meas_screen_types_str.append(item.name)
            meas_screen_types_val.append(item.value)

        meas_screen_types_str = MeasScreenTypes.get_str_list()
        self.pgi_meas_display = pg.EnumProperty(label=u"Measurement display", name=u"Measurement display",
                                                labels=meas_screen_types_str, values=meas_screen_types_val, value=0)
        self.pgi_meas_display = self.pg_trigger_settings.Append(self.pgi_meas_display)

        self.pgi_cat_trigger_actions = self.pg_trigger_settings.Append(
            pg.PropertyCategory(u"Trigger Actions", u"Trigger Actions"))
        self.pgi_pause_on_trigger = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Pause on trigger", u"Pause on trigger"))
        self.pgi_save_measurement = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Save measurement on trigger", u"Save measurement on trigger"))
        self.pgi_save_screenshot = self.pg_trigger_settings.Append(
            pg.BoolProperty(u"Save screenshot on trigger", u"Save screenshot on trigger"))

        self.pgi_save_settings = self.pg_trigger_settings.Append(
            pg.PropertyCategory(u"Save settings", u"Save settings"))
        self.pgi_base_working_dir = self.pg_trigger_settings.Append(
            pg.StringProperty(u"current base working directory", u"current base working directory"))
        self.pgi_new_wd_name = self.pg_trigger_settings.Append(
            pg.StringProperty(u"Name of new (sub) working directory", u"Name of new (sub) working directory"))
        self.pgi_meas_prefix = self.pg_trigger_settings.Append(
            pg.StringProperty(u"Measurement name prefix", u"Measurement name prefix"))
        self.pgi_comment = self.pg_trigger_settings.Append(pg.LongStringProperty(u"Comment", u"Comment"))
        self.pg_trigger_settings.SetSplitterLeft()

        self.pg_trigger_settings.SetPropertyValue(self.pgi_trigger_delay, self.ll_settings.trigger_delay)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_trigger_on_ll, self.ll_settings.trigger_on_ll_event)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_trigger_on_button, self.ll_settings.trigger_on_button)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_button_key_code, "Ctr. + T")
        self.pg_trigger_settings.DisableProperty(self.pgi_button_key_code)

        self.pg_trigger_settings.SetPropertyValue(self.pgi_sound_on_button_trigger, self.ll_settings.sound_on_button_trigger)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_sound_on_ll_trigger, self.ll_settings.sound_on_ll_trigger)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_meas_display, self.ll_settings.meas_screen_type.value)

        self.pg_trigger_settings.SetPropertyValue(self.pgi_pause_on_trigger, self.ll_settings.pause_on_trigger)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_save_measurement, self.ll_settings.save_meas_on_trigger)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_save_screenshot, self.ll_settings.save_screenshot_on_trigger)

        self.pg_trigger_settings.SetPropertyValue(self.pgi_base_working_dir, self.current_working_dir)
        self.pg_trigger_settings.DisableProperty(self.pgi_base_working_dir)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_new_wd_name, self.ll_settings.new_working_dir_name)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_meas_prefix, self.ll_settings.meas_name_prefix)
        self.pg_trigger_settings.SetPropertyValue(self.pgi_comment, self.ll_settings.comment)

        # Limit settings:
        # -----------------------------------------------------------
        self.unit_str = str(spectrum_settings.unit)
        self.stxt_hl_upper_unit.SetLabelText(self.unit_str)
        self.stxt_hl_lower_unit.SetLabelText(self.unit_str)
        self.stxt_level_unit.SetLabelText(self.unit_str)

        self.cbx_trace.Clear()
        str_trace_list = spectrum_settings.get_str_trace_list()
        self.cbx_trace.AppendItems(str_trace_list)
        last_sel_trace_str = str(self.ll_settings.trace)
        if last_sel_trace_str in str_trace_list:
            self.cbx_trace.SetStringSelection(last_sel_trace_str)
        else:
            self.cbx_trace.SetSelection(0)
            trace = self.cbx_trace.GetStringSelection()
            if trace:
                self.ll_settings.trace = SpecTraceTypes.from_string(trace)

        self.cbx_l_type.Clear()
        for item in LimitTypes:
            self.cbx_l_type.AppendItems(str(item))
        self.cbx_l_type.SetSelection(self.ll_settings.limit_type.value)

        # Limit values:
        # -----------------------------------------------------------
        self.chb_ll_type.ChangeSelection(self.ll_settings.ll_type.value)
        self.sctr_upper_lim.SetValue(self.ll_settings.upper_limit_value)
        self.sctr_lower_lim.SetValue(self.ll_settings.lower_limit_value)

        # Setup custom table
        self._selected_col = -1
        self._selected_row = -1
        self._tabl_cell_changed = False
        self.grd_ll_tbl.SetColLabelAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
        attr_float = grd.GridCellAttr()
        attr_float.SetReadOnly(False)
        attr_float.SetEditor(grd.GridCellFloatEditor(precision=6))
        attr_float.SetRenderer(grd.GridCellFloatRenderer(precision=6))

        col = 0
        self.grd_ll_tbl.SetColSize(col, 110)
        self.grd_ll_tbl.SetColLabelValue(col, u"Frequency (MHz)")
        self.grd_ll_tbl.SetColAttr(col, attr_float)

        col += 1
        self.grd_ll_tbl.SetColSize(col, 110)
        self.grd_ll_tbl.SetColLabelValue(col, u"Upper ({})".format(self.unit_str))
        self.grd_ll_tbl.SetColAttr(col, attr_float.Clone())

        col += 1
        self.grd_ll_tbl.SetColSize(col, 110)
        self.grd_ll_tbl.SetColLabelValue(col, u"Lower ({})".format(self.unit_str))
        self.grd_ll_tbl.SetColAttr(col, attr_float.Clone())

        self.btn_clear.SetBitmap(ico.img_icons_icon_delete_all.GetBitmap())
        self.btn_del_row.SetBitmap(ico.img_icons_icon_delete.GetBitmap())
        self.btn_add_row.SetBitmap(ico.img_icons_icon_add.GetBitmap())
        self.btn_update.SetBitmap(ico.img_icons_refresh.GetBitmap())

        # Add plot area:
        # ---------------------------------------------------------------------
        self.plt = TracePlotCanvasPnl(self.pnl_plot_container)
        self.szr_plot.Add(self.plt, 1, wx.EXPAND | wx.ALL, 5)
        self.count = 1

        self.Layout()
        self.Refresh()
        self.Update()

    def on_ok_button(self, event):
        """OK button clicked.

        :param event: Dialog event trigger when save button is clicked
        """
        try:
            # Trigger settings:
            # -----------------------------------------------------------
            self.ll_settings.trigger_delay = self.pg_trigger_settings.GetPropertyValue(self.pgi_trigger_delay)
            self.ll_settings.trigger_on_ll_event = self.pg_trigger_settings.GetPropertyValue(self.pgi_trigger_on_ll)
            self.ll_settings.trigger_on_button = self.pg_trigger_settings.GetPropertyValue(self.pgi_trigger_on_button)

            self.ll_settings.sound_on_button_trigger = self.pg_trigger_settings.GetPropertyValue(self.pgi_sound_on_button_trigger)
            self.ll_settings.sound_on_ll_trigger = self.pg_trigger_settings.GetPropertyValue(self.pgi_sound_on_ll_trigger)
            self.ll_settings.meas_screen_type = MeasScreenTypes(self.pg_trigger_settings.GetPropertyValue(self.pgi_meas_display))

            self.ll_settings.pause_on_trigger = self.pg_trigger_settings.GetPropertyValue(self.pgi_pause_on_trigger)
            self.ll_settings.save_meas_on_trigger = self.pg_trigger_settings.GetPropertyValue(self.pgi_save_measurement)
            self.ll_settings.save_screenshot_on_trigger = self.pg_trigger_settings.GetPropertyValue(self.pgi_save_screenshot)

            self.ll_settings.new_working_dir_name = self.pg_trigger_settings.GetPropertyValue(self.pgi_new_wd_name)
            self.ll_settings.meas_name_prefix = self.pg_trigger_settings.GetPropertyValue(self.pgi_meas_prefix)
            self.ll_settings.comment = self.pg_trigger_settings.GetPropertyValue(self.pgi_comment)

            # Limit settings:
            # -----------------------------------------------------------
            self.ll_settings.custom_tbl_csv = self.custom_tbl_csv
            self.ll_settings.custom_tbl_h5 = self.custom_tbl_h5

            fh.save_dclass_as_xml(self.ll_settings, fh.LL_LAST_SETTINGS_FILENAME)
        except:
            pass
        finally:
            event.Skip()

    def __update_plot(self):
        """Updates the plot

        :return: None
        """
        try:
            self.plt.update_settings(self.spectrum_settings, self.ll_settings)
            trace = str(self.ll_settings.trace)
            trace_lev = self.trace_data[trace]
            if trace_lev is not None:
                self.plt.update_y_values(trace, trace_lev)
            self.plt.update_plot()
        except:
            pass

    def on_show(self, event):
        """On dialog show event."""
        try:
            self.__update_table()
            self.__update_plot()
        except:
            pass
        finally:
            event.Skip()

    def on_cbx_trace(self, event):
        """Trace changed"""
        try:
            trace = self.cbx_trace.GetStringSelection()
            self.ll_settings.trace = SpecTraceTypes.from_string(trace)
            self.__update_plot()
        except:
            pass
        finally:
            event.Skip()

    def on_cbx_l_type(self, event):
        """Limit type changed"""
        self.ll_settings.limit_type = LimitTypes.from_string(self.cbx_l_type.GetStringSelection())
        self.__update_plot()
        event.Skip()


    def on_sctr_upper_lim(self, event):
        """Upper limit value changed"""
        self.ll_settings.upper_limit_value = self.sctr_upper_lim.GetValue()
        self.__update_plot()
        event.Skip()

    def on_sctr_upper_lim_mouse_wheel(self, event):
        """Upper limit value mouse wheel"""
        try:
            value = self.sctr_upper_lim.GetValue()
            inc = self.sctr_upper_lim.GetIncrement()
            wheel_rotation = event.GetWheelRotation()
            if wheel_rotation > 0:
                value += inc
            else:
                value -= inc
            self.sctr_upper_lim.SetValue(value)
        except:
            pass
        finally:
            event.Skip()

    def on_sctr_lower_lim(self, event):
        """Upper limit value changed"""
        self.ll_settings.lower_limit_value = self.sctr_lower_lim.GetValue()
        self.__update_plot()
        event.Skip()

    def on_sctr_lower_lim_mouse_wheel(self, event):
        """Lower limit value mouse wheel"""
        try:
            value = self.sctr_lower_lim.GetValue()
            inc = self.sctr_lower_lim.GetIncrement()
            wheel_rotation = event.GetWheelRotation()
            if wheel_rotation > 0:
                value += inc
            else:
                value -= inc
            self.sctr_lower_lim.SetValue(value)
        except:
            pass
        finally:
            event.Skip()

    def on_chb_ll_type(self, event):
        """Limit line type changed"""
        selection = self.chb_ll_type.GetSelection()
        self.ll_settings.ll_type = LimitLineTypes(selection)
        self.__update_plot()
        event.Skip()

    def on_grd_ll_tbl_select_cell(self, event):
        """Table cell left click"""
        try:
            self._selected_col = event.Col
            self._selected_row = event.Row

            if self._tabl_cell_changed:
                self._tabl_cell_changed = False
                self.__update_ll_settings()
        except:
            self._selected_col = -1
            self._selected_row = -1
        finally:
            event.Skip()

    def on_grd_ll_tbl_kill_fokus(self, event):
        """Table cell kill fokus"""
        try:
            self._selected_col = -1
            self._selected_row = -1
            if self._tabl_cell_changed:
                self._tabl_cell_changed = False
                self.__update_ll_settings()
        except:
            self._selected_col = -1
            self._selected_row = -1
        finally:
            event.Skip()

    def on_grd_ll_tbl_key_down(self, event):
        """Table key down event"""
        try:
            keycode = event.GetUnicodeKey()
            if keycode == wx.WXK_RETURN:
                if self._tabl_cell_changed:
                    self._tabl_cell_changed = False
                    self.__update_ll_settings()
        except:
            pass
        finally:
            event.Skip()

    def on_grd_ll_tbl_mouse_wheel(self, event):
        """Limit line table mouse wheel"""
        try:
            if self._selected_col < 0 or self._selected_row < 0:
                return

            value_str = self.grd_ll_tbl.GetCellValue(self._selected_row, self._selected_col)
            if value_str:
                value = float(value_str)
            else:
                value = 0

            wheel_rotation = event.GetWheelRotation()
            if wheel_rotation > 0:
                value += 5
            else:
                value -= 5
            self.grd_ll_tbl.SetCellValue(self._selected_row, self._selected_col, str(value))
            self._tabl_cell_changed = True
        except:
            pass
        finally:
            event.Skip()

    def on_btn_clear(self, event):
        """
        Clears the wxGrid
        """
        try:
            # Ask the user for a decision
            with wx.MessageDialog(self, 'Should the table really be cleared?', 'Clear table!',
                                  wx.YES_NO | wx.STAY_ON_TOP | wx.ICON_INFORMATION) as dlg:
                if dlg.ShowModal() == wx.ID_YES:
                    num_curr_rows = self.grd_ll_tbl.GetNumberRows()
                    if num_curr_rows > 0:
                        self.grd_ll_tbl.DeleteRows(0, num_curr_rows)
                    self.grd_ll_tbl.InsertRows(numRows=1)
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_btn_add_row(self, event):
        """
        Adds a new row at the bottom of the wxGrid
        """
        try:
            if self.grd_ll_tbl.AppendRows(numRows=1):
                num_tab_rows = self.grd_ll_tbl.GetNumberRows()
                row = num_tab_rows - 1
                self.grd_ll_tbl.SetCellValue(row, 1, "999.0")
                self.grd_ll_tbl.SetCellValue(row, 2, "-999.0")

        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_btn_del_row(self, event):
        """
        Deletes the selected row from the wxGrid
        """
        try:
            rows = self.grd_ll_tbl.GetSelectedRows()
            if rows:
                self.grd_ll_tbl.DeleteRows(rows[0])
                self.__update_ll_settings()
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_btn_update(self, event):
        """
        Updates the wxGrid
        """
        try:
            self.__update_table()
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def __update_table(self):
        """Updates the limit line table and the graph from limit line settings"""
        try:
            self.ll_settings.align_frequencies()

            self.ignore_tbl_events = True
            num_rows = len(self.ll_settings.ll_items)
            num_tab_rows = self.grd_ll_tbl.GetNumberRows()
            if num_tab_rows > 0:
                self.grd_ll_tbl.DeleteRows(0, num_tab_rows)
            self.grd_ll_tbl.InsertRows(numRows=num_rows)

            factor = 1/1e6
            row = 0
            for item in self.ll_settings.ll_items:
                freq_str = str(item.frequency*factor)
                lev_upper_str = str(item.upper_level)
                lev_lower_str = str(item.lower_level)
                self.grd_ll_tbl.SetCellValue(row, 0, freq_str)
                self.grd_ll_tbl.SetCellValue(row, 1, lev_upper_str)
                self.grd_ll_tbl.SetCellValue(row, 2, lev_lower_str)
                row += 1

            self.ignore_tbl_events = False
            self.__update_plot()
        except:
            self.ignore_tbl_events = False
            pass

    def __update_ll_settings(self) -> bool:
        """ Update ll_settings from table

        :return:
        """
        try:
            if self.ignore_tbl_events:
                return False

            self.ll_settings.ll_items.clear()

            num_tab_rows = self.grd_ll_tbl.GetNumberRows()
            for row in range(num_tab_rows):
                freq_str = self.grd_ll_tbl.GetCellValue(row, 0)
                lev_upper_str = self.grd_ll_tbl.GetCellValue(row, 1)
                lev_lower_str = self.grd_ll_tbl.GetCellValue(row, 2)
                if freq_str:
                    freq = float(freq_str)*1e6
                    if lev_upper_str:
                        lev_upper = float(lev_upper_str)
                    else:
                        lev_upper = -999.0
                    if lev_lower_str:
                        lev_lower = float(lev_lower_str)
                    else:
                        lev_lower = 999.0
                    self.ll_settings.ll_items.append(LLItem(freq, lev_upper, lev_lower))

            self.__update_plot()
            return True
        except:
            return False

    def on_grd_ll_tbl_changed(self, event):
        """Table cell has changed"""
        self.__update_ll_settings()
        event.Skip()

    def on_btn_import_h5(self, event):
        """
        Reads a HDF5 SignalShark file as custom table into the wx.grid
        """
        try:
            # Open File Dialog
            filename = self.custom_tbl_h5
            with wx.FileDialog(self, 'Import SignalShark h5 file', '', filename, 'HDF5 files (*.h5)|*.h5',
                               style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.STAY_ON_TOP) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    # Do someting with the filepath
                    filename = dlg.GetPath()
                    self.custom_tbl_h5 = filename
                else:
                    return

            with RefMeasImportDlg(self, filename, self.spectrum_settings, self.ll_settings) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    self.ll_settings.align_frequencies()
                    self.__update_table()

        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_btn_import_csv(self, event):
        """
        Reads a CSV file as custom table into the wx.grid
        """
        try:
            # Open File Dialog
            last_csv_filename = self.custom_tbl_csv
            with wx.FileDialog(self, 'Import CSV file', '', last_csv_filename, 'CSV files (*.csv)|*.csv',
                               style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.STAY_ON_TOP) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    # Do someting with the filepath
                    csv_file_path = dlg.GetPath()
                    self.custom_tbl_csv = csv_file_path
                else:
                    return

            csv_file = open(csv_file_path, "r")
            csv_reader = csv.reader(csv_file, delimiter=";")
            self.ll_settings.ll_items.clear()

            # Get number of lines of data
            lines = sum(1 for _ in csv_reader)

            # Get frequency and level data
            csv_file.seek(0)
            count = 0
            for line in csv_reader:
                count += 1
                if count == 1:
                    continue
                freq = float(line[0])*1e6
                lev_upper_str = line[1]
                if lev_upper_str:
                    lev_upper = float(lev_upper_str)
                else:
                    lev_upper = 999.0
                lev_lower_str = line[2]
                if lev_lower_str:
                    lev_lower = float(lev_lower_str)
                else:
                    lev_lower = -999.0
                self.ll_settings.ll_items.append(LLItem(freq, lev_upper, lev_lower))

            self.ll_settings.align_frequencies()
            self.__update_table()
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def on_btn_add_lev_offset(self, event):
        """Button add value clicked.

        :param event: Button event
        """
        try:
            with AddLevOffset(self, self.ll_settings.limit_type) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    # self.limit_type = LimitTypes(selected_limit_type)
                    # self.offset = self.sctr_offset.GetValue()
                    if dlg.limit_type == LimitTypes.UPPER:
                        for item in self.ll_settings.ll_items:
                            item.upper_level += dlg.offset
                    else:
                        for item in self.ll_settings.ll_items:
                            item.lower_level += dlg.offset
                    self.__update_table()
        except Exception as e:
            pass
        finally:
            event.Skip()

    def on_btn_from_marker(self, event):
        """Button by M1, M2 clicked.

        :param event: Button event
        """
        try:
            if self.spectrum_settings.fm1 <= 0 or self.spectrum_settings.fm2 <= 0:
                with wx.MessageDialog(self, 'Please enable and setup marker M1 and marker M2 to use this function!', 'No marker',
                                      wx.OK | wx.STAY_ON_TOP | wx.ICON_INFORMATION) as dlg:
                    dlg.ShowModal()
                return

            if self.spectrum_settings.fm2 > self.spectrum_settings.fm1:
                freq_m1 = self.spectrum_settings.fm1
                freq_m2 = self.spectrum_settings.fm2
            else:
                freq_m1 = self.spectrum_settings.fm2
                freq_m2 = self.spectrum_settings.fm1

            with NumValueDlg(self, self.ll_settings.last_lev_input, "Value:", str(self.spectrum_settings.unit)) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    lev = dlg.GetValue()
                    self.ll_settings.last_lev_input = lev
                    self.ll_settings.ll_items.clear()
                    self.ll_settings.ll_items.append(LLItem(freq_m1, lev, lev))
                    self.ll_settings.ll_items.append(LLItem(freq_m2, lev, lev))
                    self.__update_table()
        except Exception as e:
            pass
        finally:
            event.Skip()

    def on_btn_save_csv(self, event):
        """Save the wxGrid data to a CSV file.

        :param event: Dialog event trigger when save button is clicked
        """
        try:
            n_rows = self.grd_ll_tbl.GetNumberRows()
            if n_rows < 1:
                with wx.MessageDialog(self, 'No data to store!', 'Error',
                                      wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                    dlg.ShowModal()
                return

            # Get file name
            # -----------------------------------------------------------------------------------------
            default_file_name = dt.datetime.now().strftime('limit_line_%Y-%m-%d_%H-%M.csv')
            filename = ''
            initdir = self.custom_tbl_csv
            wildcard = 'CSV file (*.csv)|*.csv'
            with wx.FileDialog(self, message='Save limit line data to',
                               defaultDir=initdir,
                               defaultFile=default_file_name,
                               wildcard=wildcard,
                               style=wx.FD_SAVE | wx.STAY_ON_TOP) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    filename = dlg.GetPath()
                    self.custom_tbl_csv = filename
                else:
                    return

            # Write data to csv file
            # -----------------------------------------------------------------------------------------
            with open(filename, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')
                # Write csv header information
                writer.writerow(["Frequency (MHz)", "Upper ({})".format(self.unit_str),
                                 "Lower ({})".format(self.unit_str)])
                for row in range(n_rows):
                    writer.writerow([self.grd_ll_tbl.GetCellValue(row, 0),
                                     self.grd_ll_tbl.GetCellValue(row, 1),
                                     self.grd_ll_tbl.GetCellValue(row, 2)])
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()


class InfoScreenDlg(InfoScreenBaseDlg):
    """Custom dialog class for data visualization."""

    def __init__(self, parent, title: str, spectrum_settings: SpectrumSettings,
                 ll_settings: LLSettingsData, button_triggered: threading.Event):
        """Dialog class for data visualization.

        :param parent: The parent window as wx.Window.
        :param title: The dialog caption.
        :param spectrum_settings: The spectrum settings.
        :param ll_settings: Limit line settings.
        :param button_triggered: threadsafe event.
        """
        super().__init__(parent=parent)

        self.init_progress_bar(self.pgb_progress)
        self.init_dlg_buttons(self.sdbs_buttons, wx.OK)
        self.btn_ok.SetLabel("Quit")
        self.SetTitle(title)
        self.ll_settings = ll_settings
        self.button_triggered = button_triggered
        self.last_trigger_tstamp: Optional[dt.datetime] = None
        self.trigger_count = 0

        # Hotkey registration and bind. CTRL + T
        hotKeyId = 100
        self.RegisterHotKey(hotKeyId, wx.MOD_CONTROL, ord("T"))
        self.Bind(wx.EVT_HOTKEY, self.on_hotkey, id=hotKeyId)

        self.Maximize(True)

    def on_hotkey(self, event):
        """Trigger on hotkey Ctr. + T

        :param event: The hotkey event
        :return: None
        """
        self.button_triggered.set()
        event.Skip()

    def update_progress(self, evt: ProgressEvent):
        """Receives data from thread and updates progressbar and message text.

        :param evt: Event of type ProgressEvent.
        :type evt: ProgressEvent
        """
        update_layout = False

        # Handle progress
        self.update_progress_bar(evt, self.pgb_progress)

        # Handle message
        if evt.msg is not None:
            self.lbl_status.SetLabelText(evt.msg)

        # Handle data:
        if evt.data is not None:
            # data=[trigger_tstamp, trigger_source, trace_data_new]
            # Update Dialog Panel
            self._update_info_text(evt.data[0], evt.data[1])

        # Handle buttons:
        if evt.btn_style:
            self.update_btn_style(evt.btn_style)
            update_layout = True

        # Handle icons:
        if evt.icon_style:
            self.update_icon_style(self.bmp_icon, evt.icon_style)
            update_layout = True

        # Handle help text:
        if evt.help_text:
            self.help_text = evt.help_text

        # Update GUI
        if update_layout:
            self.Layout()
        # self.Refresh() # Causes this window, and all of its children recursively to be repainted.
        self.Update()  # Repaints the invalidated area

    def _update_info_text(self, trigger_tstamp: dt.datetime, trigger_source: TriggerSources):
        """Changes the message label color according to the trogger source flag.

        :param trigger_tstamp: Timestamp of trigger event..
        :param trigger_source: Flag indicating the source of the trigger event.
        :return: None
        """
        try:
            if trigger_tstamp == self.last_trigger_tstamp:
                return
            self.last_trigger_tstamp = trigger_tstamp
            if int(trigger_source) > 0:
                self.trigger_count += 1
                self.stxt_num_events.SetLabelText(str(self.trigger_count))

                trigger_flags = [flag.name for flag in TriggerSources if flag in trigger_source]
                trigger_flags_str = ", ".join(trigger_flags)
                self.stxt_last_event.SetLabelText(trigger_tstamp.strftime(trigger_flags_str + " %Y-%m-%d %H:%M:%S"))
        except:
            pass


class SpectrumPlotDlg(PlotBaseDlg):
    """Custom dialog class for data visualization."""

    def __init__(self, parent, title: str, spectrum_settings: SpectrumSettings,
                 ll_settings: LLSettingsData, button_triggered: threading.Event):
        """Dialog class for data visualization.

        :param parent: The parent window as wx.Window.
        :param title: The dialog caption.
        :param spectrum_settings: The spectrum settings.
        :param ll_settings: Limit line settings.
        :param button_triggered: threadsafe event.
        """
        super().__init__(parent=parent)

        self.init_progress_bar(self.pgb_progress)
        self.init_dlg_buttons(self.sdbs_buttons, wx.OK)
        self.btn_ok.SetLabel("Quit")
        self.SetTitle(title)
        self.spectrum_settings = spectrum_settings
        self.ll_settings = ll_settings
        self.button_triggered = button_triggered
        self.last_trigger_tstamp: Optional[dt.datetime] = None

        self.selected_trace = str(ll_settings.trace)

        # Add plot area:
        # ---------------------------------------------------------------------
        self.plt = TracePlotCanvasPnl(self.pnl_plot_container)
        self.plt.update_settings(spectrum_settings, ll_settings)
        self.szr_plot.Add(self.plt, 1, wx.EXPAND | wx.ALL, 5)
        self.count = 1

        # Hotkey registration and bind. CTRL + T
        hotKeyId = 100
        self.RegisterHotKey(hotKeyId, wx.MOD_CONTROL, ord("T"))
        self.Bind(wx.EVT_HOTKEY, self.on_hotkey, id=hotKeyId)

        self.Maximize(True)


    def on_hotkey(self, event):
        """Trigger on hotkey Ctr. + T

        :param event: The hotkey event
        :return: None
        """
        self.button_triggered.set()
        event.Skip()

    def update_progress(self, evt: ProgressEvent):
        """Receives data from thread and updates progressbar and message text.

        :param evt: Event of type ProgressEvent.
        :type evt: ProgressEvent
        """
        update_layout = False

        # Handle progress
        self.update_progress_bar(evt, self.pgb_progress)

        # Handle message
        if evt.msg is not None:
            self.lbl_status.SetLabelText(evt.msg)

        # Handle data:
        if evt.data is not None:
            # data=[trigger_tstamp, trigger_source, trace_data_new]
            # Update Dialog Panel
            self._change_label_color(evt.data[0], evt.data[1])

            # Update y values
            self.plt.update_y_values(self.selected_trace, evt.data[2])
            self.plt.update_plot()

        # Handle buttons:
        if evt.btn_style:
            self.update_btn_style(evt.btn_style)
            update_layout = True

        # Handle icons:
        if evt.icon_style:
            self.update_icon_style(self.bmp_icon, evt.icon_style)
            update_layout = True

        # Handle help text:
        if evt.help_text:
            self.help_text = evt.help_text

        # Update GUI
        if update_layout:
            self.Layout()
        self.Refresh() # Causes this window, and all of its children recursively to be repainted.
        # self.Update()  # Repaints the invalidated area

    def _change_label_color(self, trigger_tstamp: dt.datetime, trigger_source: TriggerSources):
        """Changes the message label color according to the trogger source flag.

        :param trigger_tstamp: Timestamp of trigger event..
        :param trigger_source: Flag indicating the source of the trigger event.
        :return: None
        """
        try:
            if trigger_tstamp == self.last_trigger_tstamp:
                return
            self.last_trigger_tstamp = trigger_tstamp
            if int(trigger_source) > 0:
                self.pnl_trigger_status.SetBackgroundColour((255, 168, 50))
                self.stxt_trigger_status.SetBackgroundColour((255, 168, 50))

                trigger_flags = [flag.name for flag in TriggerSources if flag in trigger_source]
                trigger_flags_str = ", ".join(trigger_flags)
                self.stxt_trigger_status.SetLabelText(trigger_tstamp.strftime("Last trigger event: " + trigger_flags_str + " %Y-%m-%d %H:%M:%S"))
            else:
                self.pnl_trigger_status.SetBackgroundColour((0, 180, 0))
                self.stxt_trigger_status.SetBackgroundColour((0, 180, 0))
                self.stxt_trigger_status.SetLabelText("Passed")
        except:
            pass

    def on_btn_save_png(self, event):
        """OK button Save as PNG clicked.

        :param event: Button event
        """
        try:
            initdir = ""
            time_stamp = dt.datetime.now()

            default_file_name = time_stamp.strftime("Screenshot_%Y-%m-%d_%H-%M-%S.png")
            wildcard = "PNG file (*.png)|*.png"
            filename = ""
            with wx.FileDialog(self, message="Save screenshot to", defaultDir=initdir,
                               defaultFile=default_file_name, wildcard=wildcard, style=wx.FD_SAVE) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    filename = dlg.GetPath()
                    self._save_screenshot(filename)
        except Exception as e:
            with wx.MessageDialog(self, str(e), 'Error',
                                  wx.OK | wx.STAY_ON_TOP | wx.ICON_ERROR) as dlg:
                dlg.ShowModal()
        finally:
            event.Skip()

    def _save_screenshot(self, filename: str) -> str:
        """Save a screenshot.
        
        :param filename: The filename of the screenshot
        :return: Error message of empty string
        """
        try:
            # screen_rect = self.GetScreenRect()
            screen_rect = self.pnl_plot_container.GetScreenRect()
            left_x, top_y = screen_rect.GetTopLeft().Get()
            right_x, bottom_y = screen_rect.GetBottomRight().Get()
            time.sleep(1.0)
            screenshot = ImageGrab.grab(bbox=(left_x, top_y, right_x, bottom_y))
            screenshot.save(filename, 'PNG')
            self.lbl_status.SetLabelText("Screenshot saved.")
            return ""
        except Exception as e:
            return str(e)
