#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2022 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################

import threading
from concurrent.futures import ThreadPoolExecutor
import time
import math as mt
import wx
import wx.grid
import winsound
import datetime as dt
from nardascripting.base.usrscriptbase import *
from nardascripting.base.signalsharkdev import *
from nardascripting.base.measthread import MeasDlgThread
import nardascripting.base.toolbox as tb
from . import filehandling as fh
from .lldata import *
from .gui.lldialogs import LLEdit, SpectrumPlotDlg, InfoScreenDlg


class LimitLines(UsrScriptBase):
    """User script class"""

    def __init__(self, main_gui, dev=SignalSharkDev()):
        """Initialization. Please leave this line unchanged"""
        super().__init__(main_gui, dev, __file__)

        # Script settings.
        # -------------------------------------------------------------------------------
        # Please adapt the following lines of code according to your script.
        self._tab_name = "nsllimitline"
        self._scr_title = "Spectrum Limit Lines V1.1"
        self._scr_description = "Acquires spectrum data, provides various options for " \
                                "limit line settings, and then shows them in a plot dialog with " \
                                "an Trigger status."
        # self._icon_path = self.script_path.joinpath("NardaScriptLauncher_example01.png")
        self._list_prio = 1
        self._nsl_executed_behavior = NSL_Executed_Behaviors.SHOW_NSL

        # Add class variables if needed
        # -------------------------------------------------------------------------------
        self.snd_lock = threading.Lock()
        self.save_lock = threading.Lock()
        self.button_triggered = threading.Event()

    def _run_script(self, args):
        """ Script main function

        This method is called when a user clicks on the corresponding script button.

        Use the dialog classes only in this part of the code.
        """
        self.ll_settings = fh.load_dclass_from_xml(LLSettingsData, fh.LL_LAST_SETTINGS_FILENAME)
        if not self.ll_settings:
            self.MessageBoxModal("Error reading limit line settings", "Settings Error",
                                 wx.OK | wx.ICON_INFORMATION)
            return

        # Connect to the device
        # -------------------------------------------------------------------------------
        if not self.signalshark.connect():
            self.MessageBoxModal("Can't connect to device.", "Connection error",
                                 wx.OK | wx.ICON_INFORMATION)
            return
        self.signalshark.scpi.check_error()

        # Check task (Spectrum / RT Spectrum):
        # -------------------------------------------------------------------------------
        task_type, task_name = self.signalshark.scpi.task.get_selected()
        if task_type != TaskTypes.SPECTRUM and task_type != TaskTypes.RT_SPECTRUM:
            self.MessageBoxModal("This script requires a Task of type \'Spectrum' or 'RT Spectrum\'",
                                 "Task", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        # Setup Spectrum View:
        # -------------------------------------------------------------------------------
        # Add a Spectrum View if it does not already exist and configure it
        exists, view_index = self.signalshark.scpi.check_add_view(ViewTypes.SPECTRUM)
        if view_index < 1:
            self.MessageBoxModal("Adding Spectrum View not possible.",
                                 "View", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        if not exists:
            self.MessageBoxModal("Please configure the newly added Spectrum View (e.g. Fstart/Fstop, RBW, ...)",
                                 "Spectrum View", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        # Setup Spectrogram View:
        # -------------------------------------------------------------------------------
        exists, view_index = self.signalshark.scpi.check_add_view(ViewTypes.SPECTROGRAM)
        if view_index < 1:
            self.MessageBoxModal("Adding Spectrogram View not possible (requires \'Option, Spectrogram\').",
                                 "View", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return


        max_sgram_lines = 10000
        meas_time = self.signalshark.scpi.spectrum.get_measurement_time()
        min_sgram_lines = int(mt.ceil((self.ll_settings.trigger_delay + 0.5) / meas_time))
        actual_sgram_lines = self.signalshark.scpi.sgram.get_frame_count()

        if min_sgram_lines > max_sgram_lines:
            self.MessageBoxModal("More than 10.000 Spectrogram lines needed. Please increase \'Meas. Time\' or "
                                 "reduce parameter (Limit Line Settings) \'time delay\'."
                                 "\r\nRule of thumb: \'Sgram Frames\' ~ (\'time delay\' + 0.5)/\'Meas. Time\'",
                                 "Meas. Time.", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        if min_sgram_lines >  actual_sgram_lines:
            self.signalshark.scpi.sgram.set_frame_count(min_sgram_lines)
            self.MessageBoxModal("Spectrogram lines increased to {}."
                                 "\r\nRule of thumb: \'Sgram Frames\' ~ (\'time delay\' + 0.5)/\'Meas. Time\'".format(min_sgram_lines),
                                 "Meas. Time.", wx.OK | wx.ICON_INFORMATION)

        # Read spectrum settings from device:
        self.spectrum_settings = SpectrumSettings.from_device(self.signalshark)

        if not self.spectrum_settings:
            self.MessageBoxModal("Error reading spectrum settings", "Settings Error",
                                 wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        self.signalshark.disconnect()

        # Prepare button trigger:
        self.button_triggered.clear()

        # Dialog for spectrum plots
        if self.ll_settings.meas_screen_type == MeasScreenTypes.INFO:
            meas_dlg = InfoScreenDlg(self.main_gui, title="Limit Line Info",
                                       spectrum_settings=self.spectrum_settings,
                                       ll_settings=self.ll_settings, button_triggered=self.button_triggered)
        else:
            meas_dlg = SpectrumPlotDlg(self.main_gui, title="Limit Line Spectrum",
                                       spectrum_settings=self.spectrum_settings,
                                       ll_settings=self.ll_settings, button_triggered=self.button_triggered)

        # Create an additional measurement thread to keep the GUI alive.
        # Please change 'MyUserScript._run_measurement' according to your class name.
        # With "args=[param1, parm2, paramn]" you can pass some optional parameters to the measurement thread.
        mthread = MeasDlgThread(main_gui=self.main_gui, dlg=meas_dlg,
                                callback=LimitLines._run_measurement,
                                args=[self.signalshark.addr,
                                      self.spectrum_settings, self.ll_settings,
                                      self.snd_lock, self.save_lock, self.button_triggered])

        # Start the measurement thread and show a visualization (progress) dialog.
        mthread.start_measurement()
        # The lines of code below will be executed after the measurement thread has done it's job.
        # ...

    @staticmethod
    def _run_measurement(stopevent, update_progress, wait_for_stopevent,
                         signalshark_addr: str, spectrum_settings: SpectrumSettings, ll_settings: LLSettingsData,
                         snd_lock: threading.Lock, save_lock: threading.Lock, button_triggered: threading.Event):
        """ Main method of the measurement thread that handles a time-consuming measurement.

        Please do NOT use any wxPython elements like a MessageBox or a Dialog here!
        They are only permitted in the GUI thread ('_run_script' method).
        Please use the "update_progress" method to display information!

        :param stopevent: Thread flag that indicates whether the procedure should be finished/cancelled
        :param update_progress: Delegate method to update a progress bar, message string and icon
        :param wait_for_stopevent: Delegate method to wait until stopevent is raised with possibility to show a message.
        :param signalshark_addr: Custom parameter to hand over the ip address of the SignalShark
        :param spectrum_settings: The spectrum settings
        :param ll_settings: The limit line settings
        :param snd_lock: Lock for sound ressource
        :param save_lock: Lock for save file ressource
        :param button_triggered: threadsafe event.
        :return: Optional custom parameter
        """
        # Setup connection to the SignalShark and display error message if connection fails.
        signalshark = SignalSharkDev(signalshark_addr)
        if not signalshark.connect():
            return wait_for_stopevent("Cannot connect to the device!", icon_style=wx.ICON_ERROR)

        old_working_dir = ""
        trigger_tstamp: Optional[dt.datetime] = None
        trigger_source: TriggerSources = TriggerSources(0)
        try:
            # Update the dialog message text and show a progress bar
            update_progress(msg=" ", show_progress=True)

            # Initialize limit line:
            limit_line = LimitLine(ll_settings)
            if not limit_line.setup_ll(spectrum_settings):
                return wait_for_stopevent("Cannot setup limit line!", icon_style=wx.ICON_ERROR)

            # Save settings
            # ---------------------------------------------------------------------------
            if ll_settings.save_meas_on_trigger or ll_settings.save_screenshot_on_trigger:
                old_working_dir = signalshark.scpi.dlogger.get_dloger_working_directory()
                base_working_dir = Path(old_working_dir)
                # Create and set new working directory and get view information
                if ll_settings.new_working_dir_name is not None and base_working_dir.stem !=  ll_settings.new_working_dir_name:
                    new_working_dir = str(base_working_dir.joinpath(ll_settings.new_working_dir_name))
                    real_path_str = signalshark.scpi.dlogger.get_directory_add(new_working_dir)

                    if not real_path_str or "NAN" in real_path_str:
                        return wait_for_stopevent("Cannot create the specified working directory!", icon_style=wx.ICON_ERROR)
                    signalshark.scpi.dlogger.set_dloger_working_directory(real_path_str)
                else:
                    old_working_dir = ""

                if not ll_settings.meas_name_prefix:
                    return wait_for_stopevent("Please define a valid measurement name prefix!", icon_style=wx.ICON_ERROR)

            last_scan_number = -1
            trace_data_list = []
            trigger_flag = False
            trigger_delay_f = float(ll_settings.trigger_delay)
            sleep_time = 1.0
            signalshark.scpi.run_cont()
            signalshark.scpi.hold()

            while True:
                if stopevent.is_set():
                    return

                # Wait until Scan Number has increased.
                scan_number = signalshark.scpi.spectrum.get_data_update(scan_number=last_scan_number)
                if scan_number > last_scan_number:
                    last_scan_number = scan_number
                else:
                    stopevent.wait(0.1)
                    continue

                # Get new trace data:
                trace_data_new = signalshark.scpi.spectrum.get_data_level(ll_settings.trace)

                # Trigger check
                if ll_settings.trigger_on_button and button_triggered.is_set():
                    button_triggered.clear()
                    trigger_source = trigger_source | TriggerSources.BUTTON

                if ll_settings.trigger_on_ll_event and limit_line.check_ll_condition(trace_data_new):
                    trigger_source = trigger_source | TriggerSources.LIMIT_LINE

                if int(trigger_source) > 0:
                    # If trigger event has occured:
                    trigger_tstamp = dt.datetime.now()

                    # Play sound
                    if ll_settings.sound_on_button_trigger or ll_settings.sound_on_ll_trigger:
                        executor = ThreadPoolExecutor(max_workers=1)
                        executor.submit(LimitLines._play_sound, snd_lock)

                    # Show trigger message:
                    msg = ""
                    if ll_settings.trigger_delay > 0:
                        msg = "Waiting for trigger delay time ({})s".format(ll_settings.trigger_delay)
                    while not update_progress(msg=msg, data=[trigger_tstamp, trigger_source, trace_data_new]):
                        stopevent.wait(0.1)
                        if stopevent.is_set():
                            return

                    # Handle trigger delay:
                    if trigger_delay_f > 0.0:
                        while True:
                            time_left = trigger_delay_f - (dt.datetime.now().timestamp() - trigger_tstamp.timestamp())
                            if time_left <= 0.0:
                                break

                            if time_left > sleep_time:
                                if stopevent.is_set():
                                    return
                                time.sleep(sleep_time)
                            else:
                                time.sleep(time_left)
                                break
                        signalshark.scpi.spectrum.get_data_update()  # Update Spectrum/Spectrogram View

                    # Save measurement/screenshot:
                    if ll_settings.save_meas_on_trigger or ll_settings.save_screenshot_on_trigger:
                        LimitLines._save(stopevent, update_progress,
                                         ll_settings, signalshark, trigger_tstamp, trigger_source)

                    # Hold display:
                    if ll_settings.pause_on_trigger:
                        return wait_for_stopevent("Trigger raised. Measurement paused.",
                                                  icon_style=wx.ICON_INFORMATION)
                    trigger_source = TriggerSources(0)
                else:
                    update_progress(data=[None, None, trace_data_new])

        except Exception as e:
            # Do some error handling
            return wait_for_stopevent(str(e), icon_style=wx.ICON_ERROR)
        finally:
            if old_working_dir is not None:
                signalshark.scpi.dlogger.set_dloger_working_directory(old_working_dir)
            if not ll_settings.pause_on_trigger:
                signalshark.scpi.run_cont()
            # Close connection.
            signalshark.disconnect()

    @staticmethod
    def _save(stopevent, update_progress,
              ll_settings: LLSettingsData, signalshark: SignalSharkDev,
              trigger_tstamp: dt.datetime, trigger_source: TriggerSources) -> bool:
        """Save measurement and screenshot

        :param stopevent: Thread flag that indicates whether the procedure should be finished/cancelled.
        :param update_progress: Delegate method to update a progress bar, message string and icon.
        :param ll_settings: Limit line settings.
        :param signalshark: The SignalShark device.
        :param trigger_tstamp: Time stamp of trigger.
        :param trigger_source: Flags indicating the trigger source.
        :return: Return True if succeeded.
        """
        try:
            # tstamp = signalshark.scpi.spectrum.get_tstamp()
            meas_filename = trigger_tstamp.strftime(ll_settings.meas_name_prefix + "_%Y-%m-%d_%H-%M-%S")

            # Ensure that the screen has been refreshed before taking the screenshot
            # ToDo?
            time.sleep(1.0)

            comment = trigger_tstamp.strftime("Date and time of the trigger event: %Y-%m-%d %H:%M:%S| ")
            trigger_flags = [flag.name for flag in TriggerSources if flag in trigger_source]
            trigger_flags_str = ", ".join(trigger_flags)
            comment += "Trigger source(s): {}| Comment: ".format(trigger_flags_str)
            comment += ll_settings.comment.replace("\r", " ").replace("\n", " " )

            if ll_settings.save_screenshot_on_trigger:
                screenshot_filename = meas_filename + ".png"
                signalshark.scpi.dlogger.save_screenshot(screenshot_filename, comment)

            if ll_settings.save_meas_on_trigger:
                signalshark.scpi.dlogger.save_dataset(meas_filename, comment)
                time.sleep(0.5)
                while not stopevent.is_set():
                    cmd_progress = signalshark.scpi.system.get_command_progress()
                    if not cmd_progress.msg:
                        break
                    update_progress(msg="Saving dataset in progress", force_update=True)
                    time.sleep(1)
                update_progress(msg=" ", force_update=True)
            return True
        except Exception as e:
            return False
        finally:
            time.sleep(1.0)

    @staticmethod
    def _play_sound(snd_lock: threading.Lock):
        """ Static method to play sound in a separate thread if an audio Trigger is set.

        :param snd_lock: Lock object to acquire or release if trigger is set (no blocking)
        :return:
        """
        if snd_lock.acquire(blocking=False):
            winsound.PlaySound('SystemHand', winsound.SND_ALIAS)
            time.sleep(1.0)
            snd_lock.release()
