#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2022 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################


import json
from pathlib import Path

from nardascripting.base.signalsharkdev import *
from nardascripting.base.usrscriptbase import *
import nardascripting.base.toolbox as tb
from . import filehandling as fh
from .lldata import *
from .gui.lldialogs import LLEdit


class LLSettings(UsrScriptBase):
    """User script class"""

    def __init__(self, main_gui, dev=SignalSharkDev()):
        """Initialization. Please leave this line unchanged"""
        super().__init__(main_gui, dev, __file__)

        # Script settings.
        # -------------------------------------------------------------------------------
        # Please adapt the following lines of code according to your script.
        self._tab_name = "nsllimitline"
        self._scr_title = "Spectrum Limit Line Settings V1.1"
        self._scr_description = "Settings for the limit line script"
        # self._icon_path = self.script_path.joinpath("NardaScriptLauncher_example01.png")
        self._list_prio = 3
        self._nsl_executed_behavior = NSL_Executed_Behaviors.SHOW_NSL

    def _run_script(self, args):
        """ Script main function

        This method is called when a user clicks on the corresponding script button.

        Use the dialog classes only in this part of the code.
        """
        # Connect to the device
        # -------------------------------------------------------------------------------
        if not self.signalshark.connect():
            self.MessageBoxModal("Can't connect to device.", "Connection error",
                                 wx.OK | wx.ICON_INFORMATION)
            return
        self.signalshark.scpi.check_error()

        # Setup Spectrum View:
        # -------------------------------------------------------------------------------
        # Add a Spectrum View if it does not already exist and configure it
        exists, view_index = self.signalshark.scpi.check_add_view(ViewTypes.SPECTRUM)
        if view_index < 1:
            self.MessageBoxModal("Adding Spectrum View not possible.",
                                 "View", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        if not exists:
            self.MessageBoxModal("Please configure the newly added Spectrum View (e.g. Fstart/Fstop, RBW, ...)",
                                 "Spectrum View", wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        # Read spectrum settings from device:
        self.spectrum_settings = SpectrumSettings.from_device(self.signalshark)

        if not self.spectrum_settings:
            self.MessageBoxModal("Error reading spectrum settings", "Settings Error",
                                 wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        trace_data = {}
        # Get trace values:
        for trace in self.spectrum_settings.trace_list:
            trace_lev = self.signalshark.scpi.spectrum.get_data_level(trace)
            trace_data[str(trace)] = trace_lev

        self.current_working_dir = self.signalshark.scpi.dlogger.get_dloger_working_directory()

        self.signalshark.disconnect()

        # Show limit line editor.
        # -------------------------------------------------------------------------------
        ll_editor = LLEdit(self.frm_dlghelper, self.spectrum_settings, trace_data, self.current_working_dir)
        self.ShowDlgModalTop(ll_editor)
        ll_editor.Destroy()  # Please do not forget to destroy the created dialog if you do not need it anymore

