#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2022 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################

import threading
from typing import List, Callable, Optional
from enum import IntFlag, auto
import wx
import wx.grid
import numpy as np
import h5py
import time
import datetime as dt
import winsound
from nardascripting.base.usrscriptbase import *
from nardascripting.base.signalsharkdev import *
import nardascripting.base.toolbox as tb

# CSV/HDF5 naming to NSL naming
TRACE_NAMES_H5 = {
    "Level_RMS": "RMS",
    "Level_pPk": "PPk",
    "Level_nPk": "MPk",
    "Level_MnR": "MnR",
    "Level_AvR": "AvR",
    "Level_MxR": "MxR",
    "Level_MxP": "MxP",
    "Level_MnP": "MnP"
}


@dataclass
class SpectrumSettings:
    """Data class for keeping track of measurement parameters."""
    # General settings
    valid: bool = False
    meas_time: float = 0.0
    comment: str = ""

    # Amplitude settings:
    trace_list: [] = None
    trace_3rd_detector: Spec3rdDetectorModes = Spec3rdDetectorModes.UNKNOWN
    trace_infinite: bool = False
    input: int = 1
    atten: float = 0.0
    unit: LevelUnits = LevelUnits.UNKNOWN
    lmax: float = 0.0
    lrange: float = 0.0

    # Frequency settings:
    fstart: float = 0.0
    fstep: float = 0.0
    fcent: float = 0.0
    bin_count: int = 0
    rbw: float = 0.0
    rbw_filter_type: RbwFilterTypes = RbwFilterTypes.NORMAL

    # Marker data:
    fm1: float = 0.0
    fm2: float = 0.0

    @property
    def fstop(self) -> float:
        """
        Calculates the stop frequency
        """
        return self.fstart + (self.fstep * self.bin_count)

    @property
    def lmin(self) -> float:
        """
        Calculates the y-axis minimum level
        """
        try:
            if self.unit == LevelUnits.A_m or self.unit == LevelUnits.V_m:
                return self.lmax/(pow(10, self.lrange/20))
            elif self.unit == LevelUnits.W_m2 or self.unit == LevelUnits.W_cm2:
                return self.lmax/(pow(10, self.lrange/10))
            else:
                return self.lmax - self.lrange
        except:
            return 0.0

    def get_flist(self, divider: float = 1.0) -> List[float]:
        """ Calculates the x-axis frequencies

        :param divider: Factor by which the frequency is divided (e.g. 1.0e6 to represent MHz)
        :type divider: float
        :return: frequency list
        :rtype: list
        """
        try:
            retvalue = []
            if self.bin_count > 0 and divider != 0.0:
                freq = self.fstart / divider
                freq_step = self.fstep / divider
                retvalue.append(freq)
                for x in range(self.bin_count - 1):
                    freq += freq_step
                    retvalue.append(freq)
            return retvalue
        except:
            return []

    def get_str_trace_list(self) -> List[str]:
        """Gets the current traces as string list.

        :return: current traces as string list
        """
        try:
            trace_list_str = [str(x) for x in self.trace_list]
            return trace_list_str
        except:
            return []

    @staticmethod
    def from_device(dev: SignalSharkDev):
        """ Creates a SpectrumSettings object and reads data from device

        :param dev: The SignalShark device
        :type dev: SignalSharkDev
        :return: SpectrumSettings
        :rtype: SpectrumSettings
        """
        retvalue = None
        try:
            retvalue = SpectrumSettings()
            if dev.connect():
                dev.scpi.check_error()
                retvalue.trace_list = dev.scpi.spectrum.get_trace_list()
                retvalue.trace_3rd_detector = dev.scpi.spectrum.get_trace_3rd_detector()
                retvalue.trace_infinite = dev.scpi.spectrum.get_trace_infinite()
                retvalue.input = dev.scpi.sense.get_input()
                retvalue.atten = dev.scpi.sense.get_attenuator()
                retvalue.unit = dev.scpi.display.get_unit()
                retvalue.lmax = dev.scpi.display.get_spectrum_lmax()
                retvalue.lrange = dev.scpi.display.get_spectrum_lrange()
                retvalue.fstart = dev.scpi.spectrum.get_data_frequency_start()
                retvalue.fstep = dev.scpi.spectrum.get_data_frequency_step()
                retvalue.bin_count = dev.scpi.spectrum.get_data_count()
                retvalue.rbw = dev.scpi.spectrum.get_rbw()
                retvalue.rbw_filter_type = dev.scpi.spectrum.get_rbw_filter_type()
                retvalue.meas_time = dev.scpi.spectrum.get_measurement_time()

                marker_list = dev.scpi.marker.get_spectrum_marker_list()
                for name, index in marker_list:
                    if name == "M1":
                        retvalue.fm1 = dev.scpi.marker.get_spectrum_frequency(index)
                    elif name == "M2":
                        retvalue.fm2 = dev.scpi.marker.get_spectrum_frequency(index)

                retvalue.valid = dev.scpi.check_no_error()
        except:
            retvalue = None
        return retvalue

    @staticmethod
    def to_device(dev: SignalSharkDev, ref_settings):
        """ Sends the SpectrumSettings to the device for configuration update.

        :param dev: The SignalShark device
        :type dev: SignalSharkDev
        :param ref_settings: SpectrumSettings
        """
        if dev.connect():
            # Copy the settings from ref to the SignalShark
            dev.scpi.spectrum.set_trace_3rd_detector(ref_settings.trace_3rd_detector)
            is_first = True
            for trace in ref_settings.trace_list:
                if is_first:
                    is_first = False
                    dev.scpi.spectrum.set_trace_enable(trace, True, True)
                else:
                    dev.scpi.spectrum.set_trace_enable(trace, True)
            dev.scpi.spectrum.set_trace_infinite(ref_settings.trace_infinite)

            dev.scpi.sense.set_input(ref_settings.input)
            dev.scpi.sense.set_attenuator(ref_settings.atten)
            dev.scpi.display.set_unit(ref_settings.unit)
            dev.scpi.display.set_spectrum_lmax(ref_settings.lmax)
            dev.scpi.display.set_spectrum_lrange(ref_settings.lrange)
            dev.scpi.spectrum.set_data_frequency_stop(ref_settings.fstop)
            dev.scpi.spectrum.set_data_frequency_start(ref_settings.fstart)
            dev.scpi.spectrum.get_rbw_filter_type(ref_settings.rbw_filter_type)
            dev.scpi.spectrum.get_rbw(ref_settings.rbw)
            dev.scpi.spectrum.get_measurement_time(ref_settings.meas_time)



# Trigger sources
class TriggerSources(IntFlag):
    LIMIT_LINE = auto()
    BUTTON = auto()
    DATE_TIME = auto()


class MeasScreenTypes(ExtEnum):
    """Enumeration of measurement screen types"""
    INFO = 0
    SPECTRUM = 1


class LimitLineTypes(ExtEnum):
    """Enumeration of limit line configurations"""
    HORIZONTAL = 0
    TABLE = 1


class LimitTypes(ExtEnum):
    """Enumeration of limit types"""
    UPPER = 0
    LOWER = 1
    BOTH = 2


@dataclass
class LLItem(object):
    """Data class for limit line items"""
    frequency: float = field(default=0.0, metadata={"XmlName": "Frequency"})
    upper_level: float = field(default=+999.0, metadata={"XmlName": "UpperLevel"})
    lower_level: float = field(default=-999.0, metadata={"XmlName": "LowerLevel"})


@dataclass
class LLSettingsData(object):
    """Data class for limit line settings"""

    # Trigger settings:
    trigger_delay: int = field(default=1, metadata={"XmlName": "TimeDelay"})
    trigger_on_ll_event: bool = field(default=True, metadata={"XmlName": "TriggerOnLL"})
    trigger_on_button: bool = field(default=True, metadata={"XmlName": "TriggerOnButton"})

    pause_on_trigger: bool = field(default=False, metadata={"XmlName": "PauseMeasOnTrigger"})
    save_meas_on_trigger: bool = field(default=True, metadata={"XmlName": "SaveMeasOnTrigger"})
    save_screenshot_on_trigger: bool = field(default=True, metadata={"XmlName": "SaveScreenshotOnTrigger"})

    sound_on_button_trigger: bool = field(default=True, metadata={"XmlName": "SoundOnButtonTrigger"})
    sound_on_ll_trigger: bool = field(default=True, metadata={"XmlName": "SoundOnLLTrigger"})

    meas_screen_type: MeasScreenTypes = field(default=MeasScreenTypes.SPECTRUM, metadata={"XmlName": "MeasScreenType"})

    new_working_dir_name: str = field(default="ProjectXY", metadata={"XmlName": "NewWorkingDirName"})
    meas_name_prefix: str = field(default="Meas", metadata={"XmlName": "MeasNamePrefix"})
    comment: str = field(default="", metadata={"XmlName": "Comment"})

    # Limit line settings:
    level_unit: LevelUnits = field(default=LevelUnits.dBm, metadata={"XmlName": "LevelUnit"})
    trace: SpecTraceTypes = field(default=SpecTraceTypes.PPk, metadata={"XmlName": "Trace"})
    limit_type: LimitTypes = field(default=LimitTypes.UPPER, metadata={"XmlName": "LimitType"})

    ll_type: LimitLineTypes = field(default=LimitLineTypes.HORIZONTAL, metadata={"XmlName": "LimitLineType"})
    upper_limit_value: float = field(default=-80.0, metadata={"XmlName": "UpperLimitValue"})
    lower_limit_value: float = field(default=-110.0, metadata={"XmlName": "LowerLimitValue"})
    custom_tbl_csv: str = field(default="", metadata={"XmlName": "CustomTableCSV"})
    custom_tbl_h5: str = field(default="", metadata={"XmlName": "CustomTableH5"})

    last_lev_input: float = field(default=0.0, metadata={"XmlName": "LastLevInput"})
    ll_items: List[LLItem] = field(default_factory=lambda: [], metadata={"XmlName": "LimitLineItems"})

    def align_frequencies(self) -> bool:
        """ Sorts frequencies.

        :return: Returns True if succeded.
        """
        try:
            # Sorts the list in ascending order of frequency.
            self.ll_items.sort(key=lambda x: x.frequency)
            return True
        except:
            return False


class LLLevels(object):
    """Container class for limit line level values"""
    def __init__(self, ll_settings: LLSettingsData, is_upper: bool = True):
        """Limit line level values

        :param ll_settings: Limit line settings
        :param is_upper: Flag indicating wheter the item is an upper limit line.
        """
        self.__ll_settings = ll_settings
        self.__is_upper = is_upper
        self.levels: List[Optional[float]] = []
        self.__num_custom_tab_values = 0
        self.__m: Optional[float] = None
        self.__b: float = 0
        self.reset(0)

    @property
    def is_upper(self) -> bool:
        """Gets a flag  indicating wheter the item is an upper limit line."""
        return self.__is_upper

    def reset(self, num_ll_values: int):
        """Resets all calculated values

        :param num_ll_values: Number of calculated limit line values.
        :return: None
        """
        self.__num_custom_tab_values = len(self.__ll_settings.ll_items)
        self.__m = None
        self.__b = 0
        self.levels.clear()
        if num_ll_values > 0:
            if self.__is_upper:
                self.levels = [-999.0] * num_ll_values
            else:
                self.levels = [999.0] * num_ll_values

    def set_lev0(self, index: int):
        """Sets the list value at the given index to the value lev0.

        :param index: The list index
        :return:
        """
        self.levels[index] = None

    def set_levn(self, index: int):
        """Sets the list value at the given index to the value levn.

        :param index: The list index
        :return:
        """
        self.levels[index] = None

    def set_calc_lev(self, index: int, f: float):
        """Sets the list value at the given index to a calculated value.

        :param index: The list index
        :param f: The actual frequency
        :return:
        """
        if self.__m is None:
            self.levels[index] = None
        else:
            self.levels[index] = (self.__m * f) + self.__b

    def calc_settings(self, custom_tab_item: LLItem, custom_tab_item1: LLItem):
        """ Calculates the formula values.

        :param custom_tab_item: Custom limit line value
        :param custom_tab_item1: Custom limit line value 1
        :return: None
        """
        try:
            f_idx = custom_tab_item.frequency
            f_idx1 = custom_tab_item1.frequency
            if self.__is_upper:
                lev_idx = custom_tab_item.upper_level
                lev_idx1 = custom_tab_item1.upper_level
            else:
                lev_idx = custom_tab_item.lower_level
                lev_idx1 = custom_tab_item1.lower_level

            if lev_idx1 is None or lev_idx is  None:
                self.__m = None
                self.__b = 0
            else:
                dy = lev_idx1 - lev_idx
                dx = f_idx1 - f_idx
                self.__m = dy / dx
                self.__b = lev_idx - (self.__m * f_idx)
        except:
            self.__m = None
            self.__b = 0


class LimitLine(object):
    """Implementation of limit line functionality"""

    def __init__(self, ll_settings: LLSettingsData):
        """ Implementation of limit line functionality

        :param ll_settings: Limit line settings
        """
        self.__ll_settings = ll_settings
        self.__ll_settings.align_frequencies()
        self.__num_custom_tab_values = len(ll_settings.ll_items)
        self.ll_frequencies: List[float] = []
        self.ll_upper = LLLevels(ll_settings, True)
        self.ll_lower = LLLevels(ll_settings, False)
        self.num_ll_values = 0
        self.__custom_tab_idx = -1
        self.__custom_tab_f_idx = 0.0
        self.__custom_tab_f_idx1 = 0.0

        # Make condition check faster by avoiding if-clauses
        self.cond_check_callback: Optional[Callable[[List[float]], bool]] = None

    def setup_ll(self, spectrum_settings: SpectrumSettings) -> bool:
        """Generates interpolated value list for faster comparison.

        :param spectrum_settings: The spectrum settings of the current measurement.
        :return: Returns True if succeeded.
        """
        try:
            if self.__ll_settings.ll_type == LimitLineTypes.HORIZONTAL:
                if self.__ll_settings.limit_type == LimitTypes.UPPER:
                    self.cond_check_callback = self.__check_hl_upper_cond
                elif self.__ll_settings.limit_type == LimitTypes.LOWER:
                    self.cond_check_callback = self.__check_hl_lower_cond
                else:
                    self.cond_check_callback = self.__check_hl_both_cond
                return True
            else:
                if self.__num_custom_tab_values < 2:
                    self.ll_frequencies.clear()
                    self.ll_upper.reset(0)
                    self.ll_lower.reset(0)
                    return False

                self.__custom_tab_idx = -1
                self.__custom_tab_f_idx = 0.0
                self.__custom_tab_f_idx1 = 0.0
                self.ll_frequencies = spectrum_settings.get_flist()
                self.num_ll_values = len(self.ll_frequencies)
                self.ll_upper.reset(self.num_ll_values)
                self.ll_lower.reset(self.num_ll_values)

                i = 0
                while i < self.num_ll_values:
                    f = self.ll_frequencies[i]
                    if f < self.__custom_tab_f_idx:
                        self.ll_upper.set_lev0(i)
                        self.ll_lower.set_lev0(i)
                        i += 1
                    elif f >= self.__custom_tab_f_idx and f <= self.__custom_tab_f_idx1:
                        self.ll_upper.set_calc_lev(i, f)
                        self.ll_lower.set_calc_lev(i, f)
                        i += 1
                    else:
                        if (self.__custom_tab_idx + 2) < self.__num_custom_tab_values:
                            self.__custom_tab_idx += 1
                            custom_item_idx = self.__ll_settings.ll_items[self.__custom_tab_idx]
                            custom_item_idx1 = self.__ll_settings.ll_items[self.__custom_tab_idx + 1]
                            self.__custom_tab_f_idx = custom_item_idx.frequency
                            self.__custom_tab_f_idx1 = custom_item_idx1.frequency
                            self.ll_upper.calc_settings(custom_item_idx, custom_item_idx1)
                            self.ll_lower.calc_settings(custom_item_idx, custom_item_idx1)
                        else:
                            self.ll_upper.set_levn(i)
                            self.ll_lower.set_levn(i)
                            i += 1

                if self.__ll_settings.limit_type == LimitTypes.UPPER:
                    self.cond_check_callback = self.__check_tab_upper_cond
                elif self.__ll_settings.limit_type == LimitTypes.LOWER:
                    self.cond_check_callback = self.__check_tab_lower_cond
                else:
                    self.cond_check_callback = self.__check_tab_both_cond
                return True
        except Exception as e:
            # Do some error handling
            return False

    def check_ll_condition(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the given limit line conditions.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            return self.cond_check_callback(level_values)
        except:
            return False

    def __check_hl_upper_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the horizontal upper limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            for value in level_values:
                if value >= self.__ll_settings.upper_limit_value:
                    return True
            return False
        except:
            return False

    def __check_hl_lower_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the horizontal lower limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            for value in level_values:
                if value <= self.__ll_settings.lower_limit_value:
                    return True
            return False
        except:
            return False

    def __check_hl_both_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the horizontal upper or lower limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            for value in level_values:
                if value >= self.__ll_settings.upper_limit_value or \
                        value <= self.__ll_settings.lower_limit_value:
                    return True
            return False
        except:
            return False

    def __check_tab_upper_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the table upper limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            num_lev_values = len(level_values)
            if num_lev_values != self.num_ll_values:
                return False
            for i in range(num_lev_values):
                if self.ll_upper.levels[i] is not None and level_values[i] >= self.ll_upper.levels[i]:
                    return True
            return False
        except:
            return False

    def __check_tab_lower_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the table lower limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            num_lev_values = len(level_values)
            if num_lev_values != self.num_ll_values:
                return False
            for i in range(num_lev_values):
                if self.ll_lower.levels[i] is not None and level_values[i] <= self.ll_lower.levels[i]:
                    return True
            return False
        except:
            return False

    def __check_tab_both_cond(self, level_values: List[float]) -> bool:
        """Checks if one of the level values hits the table upper or lower limit line.

        :param level_values:
        :return: Returns True, if limit lines condition is met.
        """
        try:
            num_lev_values = len(level_values)
            if num_lev_values != self.num_ll_values:
                return False
            for i in range(num_lev_values):
                if self.ll_upper.levels[i] is not None and level_values[i] >= self.ll_upper.levels[i]:
                    return True
                if self.ll_lower.levels[i] is not None and level_values[i] <= self.ll_lower.levels[i]:
                    return True
            return False
        except:
            return False

